#include <stdio.h>
#include <unistd.h>
#include <sys/msg.h>
#include <sys/types.h>
#include <sys/ipc.h>
#include <sys/wait.h>
#include <errno.h>
#include <string.h>
#include "mensaje.h"

#define RETARDO_A 3
#define RETARDO_B 5
#define RETARDO_C 4
#define CANT_COMPUESTOS 3

// Estructuras
int retardo_compuestos[CANT_COMPUESTOS]; // Compuestos utilizados por los agentes
int id_compuestos[CANT_COMPUESTOS]; // Id de compuestos


void inicializar_compuestos() {
	retardo_compuestos[0] = RETARDO_A; id_compuestos[0] = 'A';
	retardo_compuestos[1] = RETARDO_B; id_compuestos[1] = 'B';
	retardo_compuestos[2] = RETARDO_C; id_compuestos[2] = 'C';	
}

int main() {

	int indice;
	char id_compuesto;
	pid_t pid_compuesto;
	key_t clave_cola;
	int mqid;
	mensaje msg;

	// Inicializacion estructuras
	inicializar_compuestos();

	// Key
	clave_cola = ftok(".", 'a');

	// Cola de mensajes
	mqid = msgget(clave_cola, IPC_CREAT | IPC_EXCL | 0660);

	if (mqid < 0) {
		// Error al crear la cola
		fprintf(stderr, "Error al crear la cola de mensajes: %s\n", strerror(errno));
	}

	for (indice = 0; indice < CANT_COMPUESTOS; indice++) {
		pid_compuesto = fork();
		if (pid_compuesto == 0) break; // Proceso hijo, compuesto
	}

	if (pid_compuesto == 0) {
		int num_mensaje = 0;
		// Compuesto i-esimo
		id_compuesto = id_compuestos[indice];
		printf("Compuesto %c, comienzo\n", id_compuesto);
		while(1) {			
			if (indice == TIPO_C) { // C requiere de B para progresar
				printf("Compuesto %c, aguardo por %c \n", id_compuesto, id_compuestos[indice-1]);
				msgrcv(mqid, &msg, TAM_MENSAJE, TIPO_B, 0); // Controlar errores
			}
			// Creacion y envio de mensaje
			printf("Compuesto %c, genero y envio a la cola \n", id_compuesto);
			sleep(retardo_compuestos[indice]);			
			msg.tipo = indice;			
			msg.id_compuesto = id_compuesto;
			msg.num_compuesto = num_mensaje;
			num_mensaje++;
			msgsnd(mqid, &msg, TAM_MENSAJE, 0); // Controlar errores
		}
	} else { // Superagente
		// Espera por la finalizacion de los compuestos (x completitud, para simplificar el esquema los hijos ciclan indefinidamente)
		for (indice = 0; indice < CANT_COMPUESTOS; indice ++) {
			wait(NULL);
		}
	}

	return 0;
}
